import axios from 'axios';
import { faTwitter, faGithub } from '@fortawesome/free-brands-svg-icons';
import { faFileCode, faUndo, faTimesCircle, faCircleQuestion, faExternalLink, faCode, faTerminal, faSpinner } from '@fortawesome/free-solid-svg-icons';
import { FontAwesomeIcon } from '@fortawesome/react-fontawesome';
import { useWindowSize } from '@react-hook/window-size';
import { observer } from 'mobx-react-lite';
import React$1, { useCallback, useRef, useState, useEffect, forwardRef } from 'react';
import clsx from 'clsx';
import Editor, { loader, useMonaco } from '@monaco-editor/react';
import { flattenDiagnosticMessageText } from 'typescript';
import { motion } from 'framer-motion';
import { makeAutoObservable, runInAction, observable } from 'mobx';
import { webEndpoints, localEndpoints } from '@previewjs/api';
import { createController } from '@previewjs/iframe';
import assertNever$1, { assertNever } from 'assert-never';
import { evaluateType, dereferenceType, isValid, arrayType, objectType } from '@previewjs/type-analyzer';
import prettier from 'prettier';
import parserBabel from 'prettier/parser-babel';

function createLocalApi(url) {
  return new LocalApi(url);
}
class LocalApi {
  constructor(url) {
    if (!url.endsWith("/")) {
      url += "/";
    }
    this.url = url;
  }
  async request(...[endpoint, request]) {
    const { data } = await axios.post(`${this.url}${endpoint.path}`, request);
    if (data.kind === "error") {
      throw new Error(data.message);
    }
    return data.response;
  }
}

function createWebApi(url) {
  return new WebApi(url);
}
class WebApi {
  constructor(url) {
    if (!url.endsWith("/")) {
      url += "/";
    }
    this.url = url;
  }
  async request(...[endpoint, request]) {
    const { data } = await axios.post(`${this.url}${endpoint.path}`, request);
    return data;
  }
}

const FilePath = ({ filePath }) => {
  return /* @__PURE__ */ React$1.createElement("div", {
    className: "truncate flex flex-row items-center px-2 py-1 bg-gray-100 text-gray-800 rounded-md"
  }, /* @__PURE__ */ React$1.createElement("div", {
    className: "mr-2"
  }, /* @__PURE__ */ React$1.createElement(FontAwesomeIcon, {
    icon: faFileCode
  })), /* @__PURE__ */ React$1.createElement("div", {
    className: "text-left truncate",
    style: { direction: "rtl" }
  }, filePath));
};

const Container$2 = ({ children }) => /* @__PURE__ */ React$1.createElement("div", {
  className: "text-sm bg-gray-900 filter drop-shadow"
}, children);
const Row = ({
  children,
  className
}) => /* @__PURE__ */ React$1.createElement("div", {
  className: clsx(["flex flex-row items-center p-2", className])
}, children);
const Header = Container$2;
Header.Row = Row;

const Link = (props) => {
  const onClick = useCallback((event) => {
    if (props.href) {
      openUrl(props.href);
      event.preventDefault();
    }
    return props.onClick && props.onClick(event);
  }, [props]);
  return /* @__PURE__ */ React$1.createElement("a", {
    ...props,
    onClick
  });
};
function openUrl(url) {
  if (window.openInExternalBrowser) {
    window.openInExternalBrowser(url);
    return;
  }
  if (navigator.userAgent.includes(" Code/")) {
    window.parent.postMessage({
      command: "open-browser",
      url
    }, "*");
    return;
  }
  window.open(url, "_blank");
}

async function revalidateMonacoEditor(monaco, editor) {
  const model = editor.getModel();
  if (!model) {
    return;
  }
  const getWorker = await monaco.languages.typescript.getTypeScriptWorker();
  const worker = await getWorker(model.uri);
  const diagnostics = (await Promise.all([
    worker.getSyntacticDiagnostics(model.uri.toString()),
    worker.getSemanticDiagnostics(model.uri.toString())
  ])).reduce((a, it) => a.concat(it));
  const markers = diagnostics.map((d) => {
    const startPos = d.start || 0;
    const length = d.length || 0;
    const start = model.getPositionAt(startPos);
    const end = model.getPositionAt(startPos + length);
    return {
      severity: monaco.MarkerSeverity.Error,
      startLineNumber: start.lineNumber,
      startColumn: start.column,
      endLineNumber: end.lineNumber,
      endColumn: end.column,
      message: flattenDiagnosticMessageText(d.messageText, "\n")
    };
  });
  const owner = model.getLanguageId();
  monaco.editor.setModelMarkers(model, owner, markers);
}

loader.config({
  paths: {
    vs: "/monaco-editor/min/vs"
  }
});
const PropsEditor = ({
  documentId,
  source,
  typeDeclarationsSource,
  width,
  height,
  onUpdate,
  onReset
}) => {
  const editorRef = useRef();
  const modelRef = useRef();
  const [mounted, setMounted] = useState(false);
  const onChange = useCallback((value) => {
    if (value !== void 0 && value !== source) {
      onUpdate(value);
    }
  }, [onUpdate, source]);
  useEffect(() => {
    const editor = editorRef.current;
    if (!editor) {
      return;
    }
    editor.setPosition({
      lineNumber: 1,
      column: Number.MAX_SAFE_INTEGER
    });
    editor.setScrollTop(0);
  }, [documentId]);
  const monaco = useMonaco();
  useEffect(() => {
    if (monaco && typeDeclarationsSource) {
      const libUri = monaco.Uri.parse("ts:filename/component.d.tsx");
      const existingModel = monaco.editor.getModel(libUri);
      if (existingModel) {
        existingModel.setValue(typeDeclarationsSource);
      } else {
        monaco.editor.createModel(typeDeclarationsSource, "typescript", libUri);
      }
      if (editorRef.current) {
        revalidateMonacoEditor(monaco, editorRef.current).catch(console.warn);
      }
    }
  }, [monaco, typeDeclarationsSource]);
  useEffect(() => {
    const model = modelRef.current;
    if (!model || !source || model.getValue() === source) {
      return;
    }
    model.setValue(source);
  }, [source, mounted]);
  const beforeMount = useCallback((monaco2) => {
    monaco2.languages.typescript.typescriptDefaults.setCompilerOptions({
      target: monaco2.languages.typescript.ScriptTarget.ESNext,
      noEmit: true,
      jsx: monaco2.languages.typescript.JsxEmit.Preserve
    });
  }, []);
  const onMount = useCallback((editor, monaco2) => {
    const usageUri = monaco2.Uri.parse("file:///usage.d.tsx");
    const existingModel = monaco2.editor.getModel(usageUri);
    if (existingModel) {
      modelRef.current = existingModel;
      existingModel.setValue(source || "");
    } else {
      modelRef.current = monaco2.editor.createModel(source || "", "typescript", monaco2.Uri.parse("file:///usage.d.tsx"));
    }
    editorRef.current = editor;
    editor.setModel(modelRef.current);
    for (const model of monaco2.editor.getModels()) {
      if (model.uri.toString().startsWith("inmemory://model/")) {
        model.dispose();
      }
    }
    setMounted(true);
  }, [source]);
  return /* @__PURE__ */ React$1.createElement(React$1.Fragment, null, /* @__PURE__ */ React$1.createElement("button", {
    id: "editor-refresh-button",
    className: clsx([
      "absolute right-0 m-2 p-2 bg-gray-500 rounded-md z-50",
      onReset ? "bg-opacity-40 text-white cursor-pointer hover:bg-gray-400" : "bg-opacity-25 text-gray-500"
    ]),
    title: "Reset properties",
    disabled: !onReset,
    onClick: onReset
  }, /* @__PURE__ */ React$1.createElement(FontAwesomeIcon, {
    icon: faUndo
  })), /* @__PURE__ */ React$1.createElement(Editor, {
    width,
    height,
    theme: "vs-dark",
    defaultLanguage: "typescript",
    onChange,
    options: {
      minimap: {
        enabled: false
      },
      overviewRulerLanes: 0,
      lineNumbers: "off",
      folding: false,
      renderLineHighlight: "none",
      lineDecorationsWidth: 8,
      lineNumbersMinChars: 0,
      scrollbar: {
        vertical: "hidden"
      },
      padding: {
        top: 8,
        bottom: 8
      },
      fontSize: 14
    },
    beforeMount,
    onMount
  }), /* @__PURE__ */ React$1.createElement("div", {
    id: "editor-content",
    "data-value": source,
    style: { display: "none" }
  }), mounted && /* @__PURE__ */ React$1.createElement("div", {
    id: "editor-ready"
  }));
};

const SmallLogo = (props) => {
  return /* @__PURE__ */ React.createElement(Link, {
    className: "inline-flex items-center text-gray-50 ml-2 pl-0.5 pr-1.5 py-0.5 text-sm font-semibold rounded whitespace-nowrap",
    target: "_blank",
    href: props.href,
    title: props.title
  }, /* @__PURE__ */ React.createElement("svg", {
    width: "512px",
    height: "512px",
    viewBox: "0 0 512 512",
    xmlns: "http://www.w3.org/2000/svg",
    xmlnsXlink: "http://www.w3.org/1999/xlink",
    className: "h-6 w-6 mr-1"
  }, /* @__PURE__ */ React.createElement("defs", null, /* @__PURE__ */ React.createElement("path", {
    d: "M178.998,315 L179,439 L105,439 L105,46 L240,46 L240,46.004 L241.724203,46.0180192 C314.980631,47.2057605 374,106.960524 374,180.5 C374,254.281294 314.591849,314.187071 241.001081,314.991796 L241,315 L178.998,315 Z M240.5,109 C200.459356,109 168,141.459356 168,181.5 C168,221.540644 200.459356,254 240.5,254 C280.540644,254 313,221.540644 313,181.5 C313,141.459356 280.540644,109 240.5,109 Z",
    id: "path-1"
  }), /* @__PURE__ */ React.createElement("filter", {
    x: "-9.3%",
    y: "-10.2%",
    width: "137.2%",
    height: "122.9%",
    filterUnits: "objectBoundingBox",
    id: "filter-2"
  }, /* @__PURE__ */ React.createElement("feOffset", {
    dx: "30",
    dy: "5",
    in: "SourceAlpha",
    result: "shadowOffsetOuter1"
  }), /* @__PURE__ */ React.createElement("feGaussianBlur", {
    stdDeviation: "10",
    in: "shadowOffsetOuter1",
    result: "shadowBlurOuter1"
  }), /* @__PURE__ */ React.createElement("feColorMatrix", {
    values: "0 0 0 0 0   0 0 0 0 0   0 0 0 0 0  0 0 0 0.253425045 0",
    type: "matrix",
    in: "shadowBlurOuter1"
  }))), /* @__PURE__ */ React.createElement("g", {
    id: "logo",
    stroke: "none",
    "stroke-width": "1",
    fill: "none",
    "fill-rule": "evenodd"
  }, /* @__PURE__ */ React.createElement("g", {
    id: "Group",
    transform: "translate(17.000000, 17.000000)"
  }, /* @__PURE__ */ React.createElement("rect", {
    id: "Rectangle",
    fill: "#3B83F6",
    x: "4",
    y: "4",
    width: "470",
    height: "470",
    rx: "69"
  }), /* @__PURE__ */ React.createElement("g", {
    id: "Oval"
  }, /* @__PURE__ */ React.createElement("use", {
    fill: "black",
    "fill-opacity": "1",
    filter: "url(#filter-2)",
    xlinkHref: "#path-1"
  }), /* @__PURE__ */ React.createElement("use", {
    fill: "#FFFFFF",
    "fill-rule": "evenodd",
    xlinkHref: "#path-1"
  }))))), props.label);
};

const TabbedPanel = (props) => {
  const [currentTabKey, setCurrentTabKey] = useState(props.defaultTabKey);
  const currentTab = props.tabs.find((tab) => tab.key === currentTabKey);
  return /* @__PURE__ */ React$1.createElement("div", null, /* @__PURE__ */ React$1.createElement("div", {
    className: "flex flex-row flex-wrap items-center bg-white shadow-inner"
  }, props.tabs.map((tab) => /* @__PURE__ */ React$1.createElement("button", {
    key: tab.key,
    title: tab.label,
    onClick: () => {
      setCurrentTabKey(tab.key === currentTabKey ? null : tab.key);
    },
    className: clsx([
      "panel-tab",
      "px-4 py-3 self-stretch text-sm font-medium flex flex-row items-center",
      tab.key === currentTabKey ? "panel-tab-selected bg-gray-200 text-gray-900" : "text-gray-400"
    ])
  }, /* @__PURE__ */ React$1.createElement(FontAwesomeIcon, {
    icon: tab.icon
  }), /* @__PURE__ */ React$1.createElement("div", {
    className: clsx(["ml-2"])
  }, tab.label), tab.notificationCount > 0 && /* @__PURE__ */ React$1.createElement("span", {
    className: "notification-count bg-red-600 text-white p-2 rounded-full inline-flex items-center justify-center h-5 min-w-5 ml-1 text-xs"
  }, tab.notificationCount)))), currentTab && /* @__PURE__ */ React$1.createElement("div", {
    className: "flex flex-col bg-gray-200",
    style: { height: props.height }
  }, currentTab.panel));
};

function decodeComponentId(componentId) {
  const colonPosition = componentId.lastIndexOf(":");
  if (colonPosition === -1) {
    return {
      currentFilePath: componentId,
      component: null
    };
  }
  const currentFilePath = componentId.slice(0, colonPosition);
  const componentKey = componentId.slice(colonPosition + 1);
  const slashPosition = componentKey.lastIndexOf("/");
  if (slashPosition === -1) {
    return {
      currentFilePath,
      component: {
        filePath: currentFilePath,
        name: componentKey
      }
    };
  }
  const dirPath = currentFilePath.slice(0, currentFilePath.lastIndexOf("/"));
  const siblingFileName = componentKey.slice(0, slashPosition);
  const componentName = componentKey.slice(slashPosition + 1);
  return {
    currentFilePath,
    component: {
      filePath: `${dirPath}/${siblingFileName}`,
      name: componentName
    }
  };
}

const Container$1 = ({
  children
}) => {
  return /* @__PURE__ */ React$1.createElement("div", {
    className: "absolute overflow-hidden bottom-0 right-0 z-50"
  }, children);
};

const Item$1 = (props) => {
  return /* @__PURE__ */ React$1.createElement(motion.div, {
    animate: { x: "100%" },
    transition: { delay: 3 },
    onAnimationComplete: props.onAnimationComplete
  }, /* @__PURE__ */ React$1.createElement(motion.div, {
    animate: { x: 0, marginBottom: 0 },
    initial: { x: "100%", marginBottom: -100 },
    transition: { type: "tween" }
  }, /* @__PURE__ */ React$1.createElement("div", {
    className: "action-log bg-blue-700 text-blue-100 bg-opacity-80 p-2 m-2 border-2 border-blue-300"
  }, props.children)));
};

const PopOver = {
  Container: Container$1,
  Item: Item$1
};

const ActionLogs = observer(({ state }) => {
  return /* @__PURE__ */ React$1.createElement(PopOver.Container, null, state.logs.map((item) => /* @__PURE__ */ React$1.createElement(PopOver.Item, {
    key: item.key,
    onAnimationComplete: item.remove
  }, item.action.type === "fn" ? /* @__PURE__ */ React$1.createElement(React$1.Fragment, null, "Function prop invoked: ", /* @__PURE__ */ React$1.createElement("b", null, item.action.path)) : /* @__PURE__ */ React$1.createElement(React$1.Fragment, null, "Redirect prevented:", " ", /* @__PURE__ */ React$1.createElement(Link, {
    href: item.action.path
  }, item.action.path)), item.count > 1 && ` (x${item.count})`)));
});

class ActionLogsState {
  constructor() {
    this.logs = [];
    makeAutoObservable(this);
  }
  onAction(action) {
    const key = `${Date.now()}-${action.type}-${action.path}`;
    const existingActionLog = this.logs.find((a) => a.action.path === action.path);
    if (existingActionLog) {
      existingActionLog.count += 1;
    } else {
      const item = {
        key,
        action,
        count: 1,
        remove: () => this.removeItem(item)
      };
      this.logs.push(item);
    }
  }
  removeItem(item) {
    this.logs = this.logs.filter((a) => a.key !== item.key);
  }
}

const Container = (props) => {
  const scrollableContainerRef = useRef(null);
  useEffect(() => {
    if (!scrollableContainerRef.current) {
      return;
    }
    scrollableContainerRef.current.scroll({
      behavior: "smooth",
      top: scrollableContainerRef.current.scrollHeight
    });
  });
  return /* @__PURE__ */ React$1.createElement("div", {
    id: "console-container",
    className: "flex-grow flex flex-col min-h-0"
  }, /* @__PURE__ */ React$1.createElement("div", {
    ref: scrollableContainerRef,
    className: "flex flex-col overflow-y-auto"
  }, props.children), props.children.length > 0 && /* @__PURE__ */ React$1.createElement("button", {
    id: "clear-console-button",
    className: "p-2 text-gray-400 border-t-2 border-gray-100 hover:text-gray-600",
    onClick: props.onClear
  }, /* @__PURE__ */ React$1.createElement(FontAwesomeIcon, {
    icon: faTimesCircle
  }), " Clear all"));
};

const dateFormat = new Intl.DateTimeFormat("default", {
  hour: "numeric",
  minute: "numeric",
  second: "numeric"
});
const Item = (props) => {
  return /* @__PURE__ */ React$1.createElement("code", {
    className: clsx([
      "console-item",
      `console-item-${props.level}`,
      "block p-2 text-xs whitespace-pre-wrap",
      (props.level === "log" || props.level === "info") && "bg-blue-50 odd:bg-blue-100 text-blue-400",
      props.level === "warn" && "bg-orange-50 odd:bg-orange-100 text-orange-400",
      props.level === "error" && "bg-red-50 odd:bg-red-100 text-red-600"
    ])
  }, "[", dateFormat.format(props.timestamp), "] ", props.message, props.suggestion && /* @__PURE__ */ React$1.createElement(Link, {
    className: "block mt-2 p-2 bg-white shadow text-blue-800 font-sans",
    href: props.suggestion.url
  }, /* @__PURE__ */ React$1.createElement(FontAwesomeIcon, {
    icon: faCircleQuestion,
    className: "mr-2"
  }), props.suggestion.message));
};

const ConsoleLogs = {
  Container,
  Item
};

const ConsolePanel = observer(({ state }) => {
  return /* @__PURE__ */ React$1.createElement(ConsoleLogs.Container, {
    onClear: state.onClear.bind(state)
  }, state.logs.map((log) => /* @__PURE__ */ React$1.createElement(ConsoleLogs.Item, {
    key: `${log.timestamp}-${log.message}`,
    ...log
  })));
});

class ConsolePanelState {
  constructor() {
    this.logs = [];
    makeAutoObservable(this);
  }
  onConsoleMessage(event) {
    if (this.logs.find((log) => event.level === log.level && event.message === log.message && Math.abs(event.timestamp - log.timestamp) < 1e3)) {
      return;
    }
    this.logs.push({
      ...event,
      suggestion: generateSuggestion(event.message)
    });
  }
  onClear() {
    this.logs = [];
  }
  get unreadCount() {
    return this.logs.filter((message) => isNotifiable(message.level)).length;
  }
}
function isNotifiable(level) {
  switch (level) {
    case "warn":
    case "error":
      return true;
    default:
      return false;
  }
}
function generateSuggestion(message) {
  if (message.includes(`Failed to resolve import `)) {
    const match = message.match(/Failed to resolve import "((@[a-zA-Z0-9\\-]+\/[a-zA-Z0-9\\-]+)|[a-zA-Z0-9\\-]+)"/);
    const url = "https://previewjs.com/docs/config/aliases";
    return {
      message: match ? `Perhaps you need to install "${match[1]}" or configure aliases? See ${url}` : "Perhaps you need to install a peer dependency or configure aliases? See ${url}",
      url
    };
  }
  if (message.includes(".svg?import")) {
    const url = "https://previewjs.com/docs/config/svgr";
    return {
      message: `Read ${url} to configure SVGR for your project.`,
      url
    };
  }
  return void 0;
}

const LinkBanner = (props) => {
  const dismiss = useCallback((e) => {
    e.stopPropagation();
    e.preventDefault();
    if (props.onDismiss) {
      props.onDismiss();
    }
  }, [props.onDismiss]);
  return /* @__PURE__ */ React$1.createElement(Link, {
    className: clsx([
      "flex flex-row items-center px-2 py-1",
      props.type === "warn" && "bg-red-300",
      props.type === "info" && "bg-blue-300"
    ]),
    href: props.href
  }, /* @__PURE__ */ React$1.createElement("div", {
    className: "m-2"
  }, props.message), /* @__PURE__ */ React$1.createElement("div", {
    className: "flex-grow"
  }), /* @__PURE__ */ React$1.createElement("button", {
    className: clsx([
      "px-2 py-1 rounded-md font-bold",
      props.type === "warn" && "bg-red-50",
      props.type === "info" && "bg-red-50"
    ])
  }, props.buttonLabel), props.onDismiss && /* @__PURE__ */ React$1.createElement("button", {
    className: "ml-1 px-2 py-1",
    onClick: dismiss
  }, "Dismiss"));
};

const UpdateBanner = observer(({ state: { update, onUpdateDismissed } }) => {
  if (!update) {
    return null;
  }
  return /* @__PURE__ */ React$1.createElement(LinkBanner, {
    type: update.required ? "warn" : "info",
    href: update.url,
    message: update.bannerMessage,
    buttonLabel: "Update now",
    onDismiss: update.required ? void 0 : onUpdateDismissed
  });
});

class UpdateBannerState {
  constructor(webApi, persistedStateController) {
    this.webApi = webApi;
    this.persistedStateController = persistedStateController;
    this.checkVersionResponse = null;
    this.onUpdateDismissed = async () => {
      await this.persistedStateController.update({
        updateDismissed: {
          timestamp: Date.now()
        }
      });
    };
    makeAutoObservable(this);
  }
  async start(appInfo) {
    try {
      const checkVersionResponse = await this.webApi.request(webEndpoints.CheckVersion, {
        appInfo
      });
      runInAction(() => {
        this.checkVersionResponse = checkVersionResponse;
      });
    } catch (e) {
      console.warn(e);
    }
  }
  get update() {
    const update = this.checkVersionResponse?.update;
    const persistedState = this.persistedStateController.state;
    if (!update?.available || !persistedState) {
      return null;
    }
    const dismissedAt = persistedState.updateDismissed?.timestamp;
    if (!update.required && dismissedAt && Date.now() < dismissedAt + 24 * 60 * 60 * 1e3) {
      return null;
    }
    return update;
  }
}

const Preview = observer(({
  state,
  appLabel,
  headerAddon,
  subheader,
  footer,
  panelTabs,
  viewport
}) => {
  const [width, height] = useWindowSize();
  const panelHeight = height * 0.3;
  if (!state.reachable) {
    return /* @__PURE__ */ React$1.createElement("div", {
      className: "h-screen bg-gray-700 text-gray-100 p-2 grid place-items-center text-lg"
    }, /* @__PURE__ */ React$1.createElement("div", {
      id: "app-error",
      className: "p-4 rounded-lg text-red-800 bg-red-50 filter drop-shadow-lg"
    }, "Server disconnected. Is Preview.js still running?"));
  }
  return /* @__PURE__ */ React$1.createElement("div", {
    className: "flex flex-col h-screen overflow-hidden bg-white"
  }, /* @__PURE__ */ React$1.createElement(ActionLogs, {
    state: state.actionLogs
  }), /* @__PURE__ */ React$1.createElement(Header, null, /* @__PURE__ */ React$1.createElement(Header.Row, null, headerAddon, state.component?.componentId && /* @__PURE__ */ React$1.createElement(React$1.Fragment, null, /* @__PURE__ */ React$1.createElement(FilePath, {
    key: "file",
    filePath: decodeComponentId(state.component.componentId).currentFilePath
  }), /* @__PURE__ */ React$1.createElement(Link, {
    href: document.location.href,
    target: "_blank",
    title: "Open in new tab",
    className: "text-gray-500 hover:text-gray-200 ml-2 text-lg"
  }, /* @__PURE__ */ React$1.createElement(FontAwesomeIcon, {
    icon: faExternalLink,
    fixedWidth: true
  }))), /* @__PURE__ */ React$1.createElement("div", {
    className: "flex-grow"
  }), /* @__PURE__ */ React$1.createElement(SmallLogo, {
    href: "https://previewjs.com/docs",
    label: appLabel,
    title: state.appInfo?.version && `v${state.appInfo.version}`
  }), /* @__PURE__ */ React$1.createElement(Link, {
    className: "ml-2 text-xl text-[#1DA1F2] hover:text-white",
    href: "https://twitter.com/previewjs",
    target: "_blank",
    title: "Follow Preview.js on Twitter"
  }, /* @__PURE__ */ React$1.createElement(FontAwesomeIcon, {
    icon: faTwitter,
    fixedWidth: true
  })), /* @__PURE__ */ React$1.createElement(Link, {
    className: "ml-2 text-xl bg-[#333] text-white rounded-md hover:bg-white hover:text-[#333]",
    href: "https://github.com/fwouts/previewjs",
    target: "_blank",
    title: "Star Preview.js on GitHub"
  }, /* @__PURE__ */ React$1.createElement(FontAwesomeIcon, {
    icon: faGithub,
    fixedWidth: true
  }))), subheader && /* @__PURE__ */ React$1.createElement(Header.Row, {
    className: "bg-gray-100"
  }, subheader)), /* @__PURE__ */ React$1.createElement(UpdateBanner, {
    state: state.updateBanner
  }), state.component ? viewport : /* @__PURE__ */ React$1.createElement("div", {
    id: "no-selection",
    className: "flex-grow bg-gray-700 text-gray-100 p-2 grid place-items-center text-lg"
  }, "Please select a component to preview."), /* @__PURE__ */ React$1.createElement(TabbedPanel, {
    defaultTabKey: "props",
    tabs: [
      ...state.component?.variantKey === null || state.component?.variantKey === "custom" ? [
        {
          label: "Properties",
          key: "props",
          icon: faCode,
          notificationCount: 0,
          panel: /* @__PURE__ */ React$1.createElement(PropsEditor, {
            documentId: state.component.componentId,
            height,
            width,
            onUpdate: state.updateProps.bind(state),
            onReset: state.component?.details && state.component.details.props.isDefaultInvocationSource ? void 0 : state.resetProps.bind(state),
            source: state.component.details?.props.invocationSource,
            typeDeclarationsSource: state.component.details?.props.typeDeclarations
          })
        }
      ] : [],
      ...state.component ? [
        {
          label: "Console",
          key: "console",
          icon: faTerminal,
          notificationCount: state.consoleLogs.unreadCount,
          panel: /* @__PURE__ */ React$1.createElement(ConsolePanel, {
            state: state.consoleLogs
          })
        }
      ] : [],
      ...panelTabs || []
    ],
    height: panelHeight
  }), footer);
});

const Pill = forwardRef((props, ref) => {
  return /* @__PURE__ */ React$1.createElement("div", {
    ref,
    className: "flex flex-row shrink-0 items-stretch cursor-pointer rounded-full overflow-hidden bg-white border-2 border-gray-300 text-gray-800",
    onClick: props.onClick
  }, /* @__PURE__ */ React$1.createElement("span", {
    id: "component-label",
    className: "m-1 px-2"
  }, props.loading ? /* @__PURE__ */ React$1.createElement(FontAwesomeIcon, {
    className: "mr-2 animate-spin",
    icon: faSpinner,
    fixedWidth: true
  }) : /* @__PURE__ */ React$1.createElement(FontAwesomeIcon, {
    className: "mr-2",
    icon: props.icon,
    fixedWidth: true
  }), props.label), props.buttons?.map((v) => {
    const selected = props.selectedButtonKey === v.key;
    return /* @__PURE__ */ React$1.createElement("div", {
      key: v.key,
      className: clsx([
        "button",
        "flex flex-row items-center px-3 font-light",
        selected ? "text-gray-100 bg-gray-600" : "text-gray-800 hover:bg-gray-200"
      ]),
      id: selected ? "selected-button" : void 0,
      onClick: (e) => {
        if (props.onButtonClicked) {
          props.onButtonClicked(v.key);
        }
        e.stopPropagation();
      }
    }, v.label);
  }));
});

const Selection = observer(forwardRef(({
  state,
  icon = faCode
}, ref) => {
  if (!state.component) {
    return null;
  }
  return /* @__PURE__ */ React.createElement(Pill, {
    ref,
    key: "component",
    icon,
    label: state.component.name,
    loading: !state.component.details?.variants,
    buttons: state.component.details?.variants?.filter((v) => !v.isEditorDriven) || void 0,
    selectedButtonKey: state.component.variantKey,
    onClick: () => state.setVariant("custom"),
    onButtonClicked: state.setVariant.bind(state)
  });
}));

class PersistedStateController {
  constructor(localApi) {
    this.localApi = localApi;
    this.persistedState = null;
    makeAutoObservable(this);
  }
  async start() {
    const persistedState = await this.localApi.request(localEndpoints.GetState);
    runInAction(() => {
      this.persistedState = persistedState;
    });
  }
  get state() {
    return this.persistedState;
  }
  async update(stateUpdate) {
    runInAction(() => {
      this.persistedState = {
        ...this.persistedState,
        ...stateUpdate
      };
    });
    await this.localApi.request(localEndpoints.UpdateState, stateUpdate);
  }
}

function extractFunctionKeys(propsType, types) {
  propsType = resolveType(propsType);
  if (propsType.kind !== "object") {
    return [];
  }
  return Object.entries(propsType.fields).filter(([propertyName, propertyType]) => resolveType(propertyType).kind === "function").map(([propertyName]) => propertyName);
  function resolveType(type) {
    if (type.kind === "name") {
      const resolved = types[type.name];
      if (resolved) {
        return evaluateType(resolved, []);
      }
    }
    return type;
  }
}

const RESERVED_IDENTIFIERS = /* @__PURE__ */ new Set(["arguments", "get", "set"]);
function isValidPropName(name) {
  return /^[a-zA-Z_$][a-zA-Z_$0-9]*$/.test(name) && !RESERVED_IDENTIFIERS.has(name);
}

function array(items) {
  return {
    kind: "array",
    items
  };
}
const EMPTY_ARRAY = array([]);
function boolean(value) {
  return {
    kind: "boolean",
    value
  };
}
const FALSE = boolean(false);
const TRUE = boolean(true);
function fn(returnValue) {
  return {
    kind: "function",
    returnValue
  };
}
function map(values) {
  return {
    kind: "map",
    values
  };
}
const NULL = { kind: "null" };
function number(value) {
  return {
    kind: "number",
    value
  };
}
function object(entries) {
  return {
    kind: "object",
    entries
  };
}
function promise(value) {
  return {
    kind: "promise",
    value
  };
}
const EMPTY_OBJECT = object([]);
function set(values) {
  return {
    kind: "set",
    values
  };
}
function string(value) {
  return {
    kind: "string",
    value
  };
}
const UNDEFINED = { kind: "undefined" };

function generateSerializableValue(type, collected, isFunctionReturnValue = false) {
  return _generateSerializableValue(type, collected, [], [], isFunctionReturnValue);
}
function _generateSerializableValue(type, collected, path, rejectTypeNames, isFunctionReturnValue) {
  let encounteredAliases;
  [type, encounteredAliases] = dereferenceType(type, collected, rejectTypeNames);
  rejectTypeNames = [...rejectTypeNames, ...encounteredAliases];
  switch (type.kind) {
    case "any":
    case "unknown":
    case "never":
    case "void":
      return UNDEFINED;
    case "null":
      return NULL;
    case "boolean":
      return FALSE;
    case "string":
    case "node":
      return string(path.length === 0 ? "node" : path.join("."));
    case "number":
      return number(100);
    case "literal":
      if (typeof type.value === "number") {
        return number(type.value);
      } else if (typeof type.value === "string") {
        return string(type.value);
      } else {
        return type.value ? TRUE : FALSE;
      }
    case "enum":
      const value = Object.values(type.options)[0];
      if (typeof value === "number") {
        return number(value);
      } else {
        return string(value || "unknown");
      }
    case "array":
      return generateArrayValue(type, collected, path, rejectTypeNames, isFunctionReturnValue);
    case "set": {
      return set(generateArrayValue(arrayType(type.items), collected, path, rejectTypeNames, isFunctionReturnValue));
    }
    case "object": {
      const entries = [];
      for (const [propName, propType] of Object.entries(type.fields)) {
        const propValue = _generateSerializableValue(propType, collected, [...path, propName], rejectTypeNames, isFunctionReturnValue);
        if (propValue.kind === "undefined") {
          continue;
        }
        if (!isValidPropName(propName)) {
          continue;
        }
        entries.push({
          key: string(propName),
          value: propValue
        });
      }
      return object(entries);
    }
    case "map":
      return map(EMPTY_OBJECT);
    case "record":
      return EMPTY_OBJECT;
    case "union":
      if (isValid(type, collected, void 0)) {
        return UNDEFINED;
      }
      if (isValid(type, collected, null)) {
        return NULL;
      }
      if (isValid(type, collected, false)) {
        return FALSE;
      }
      return _generateSerializableValue(type.types[0], collected, path, rejectTypeNames, isFunctionReturnValue);
    case "intersection":
      return _generateSerializableValue(type.types[0], collected, path, rejectTypeNames, isFunctionReturnValue);
    case "function":
      const returnValue = isFunctionReturnValue ? UNDEFINED : _generateSerializableValue(type.returnType, collected, path, rejectTypeNames, true);
      return fn(returnValue);
    case "optional":
      return UNDEFINED;
    case "promise": {
      return promise({
        type: "reject",
        message: null
      });
    }
    case "name":
      return _generateSerializableValue(type, collected, path, rejectTypeNames, isFunctionReturnValue);
    default:
      throw assertNever(type);
  }
}
function generateArrayValue(type, collected, path, rejectTypeNames, isFunctionReturnValue) {
  if (isFunctionReturnValue) {
    return EMPTY_ARRAY;
  }
  const itemValue = _generateSerializableValue(type.items, collected, path, rejectTypeNames, isFunctionReturnValue);
  if (itemValue.kind === "undefined" || itemValue.kind === "object" && Object.keys(itemValue.entries).length === 0) {
    return EMPTY_ARRAY;
  }
  return array([itemValue]);
}

function serializableValueToJavaScript(value) {
  switch (value.kind) {
    case "array":
      return `[${value.items.map((item) => serializableValueToJavaScript(item)).join(", ")}]`;
    case "boolean":
      return value.value ? "true" : "false";
    case "function":
      return `() => ${value.returnValue.kind === "undefined" ? "{}" : `(${serializableValueToJavaScript(value.returnValue)})`}`;
    case "map":
      return `new Map(Object.entries(${serializableValueToJavaScript(value.values)}))`;
    case "null":
      return "null";
    case "number":
      return value.value.toString(10);
    case "object": {
      if (Object.entries(value.entries).length === 0) {
        return "{}";
      }
      let text = "";
      text += "{\n";
      for (const entry of value.entries) {
        text += `${entry.key.kind === "string" ? JSON.stringify(entry.key.value) : `[${serializableValueToJavaScript(entry.key)}]`}: ${serializableValueToJavaScript(entry.value)},
`;
      }
      text += "\n}";
      return text;
    }
    case "promise":
      if (value.value.type === "reject") {
        return `Promise.reject(${value.value.message === null ? "" : JSON.stringify(value.value.message)})`;
      } else {
        return `Promise.resolve(${serializableValueToJavaScript(value.value.value)})`;
      }
    case "set":
      return `new Set(${serializableValueToJavaScript(value.values)})`;
    case "string":
      return JSON.stringify(value.value);
    case "undefined":
      return "undefined";
    case "unknown":
      return "{}";
    default:
      throw assertNever$1(value);
  }
}

function generateDefaultProps(propsType, types) {
  propsType = resolveType(propsType);
  if (propsType.kind !== "object") {
    return "{}";
  }
  const propKeys = /* @__PURE__ */ new Set();
  let text = "";
  text += "{";
  for (const [propertyName, propertyType] of Object.entries(propsType.fields)) {
    const resolvedPropertyType = resolveType(propertyType);
    if (resolvedPropertyType.kind === "function") {
      propKeys.add(propertyName);
      const returnType = resolvedPropertyType.returnType;
      text += `"${propertyName.replace(/"/g, '\\"')}": fn("${propertyName.replace(/"/g, '\\"')}", ${serializableValueToJavaScript(generateSerializableValue(returnType, types, true))}),`;
    }
  }
  text += "}";
  return text;
  function resolveType(type) {
    if (type.kind === "name") {
      const resolved = types[type.name];
      if (resolved) {
        return evaluateType(resolved, []);
      }
    }
    return type;
  }
}

function generateInvocation(propsType, providedKeys, collected) {
  const providedKeySet = new Set(providedKeys);
  [propsType] = dereferenceType(propsType, collected, []);
  if (propsType.kind === "object") {
    propsType = objectType(Object.fromEntries(Object.entries(propsType.fields).filter(([fieldName]) => !providedKeySet.has(fieldName))));
  }
  const value = generateSerializableValue(propsType, collected);
  let valueSource = serializableValueToJavaScript(value);
  if (valueSource === "undefined") {
    valueSource = "{}";
  }
  const source = `properties = ${valueSource}`;
  return prettier.format(source, {
    parser: "babel",
    plugins: [parserBabel],
    filepath: "component.js"
  }).trim();
}

function generateTypeDeclarations(name, type, argKeys, collected) {
  const typeNameMapping = {};
  const usedTypes = /* @__PURE__ */ new Set();
  const declaredTypes = /* @__PURE__ */ new Set();
  let hasProperType;
  let propsTypeName;
  if (type.kind === "name") {
    hasProperType = true;
    propsTypeName = type.name;
    usedTypes.add(type.name);
  } else {
    hasProperType = false;
    propsTypeName = `:${name}Props`;
  }
  const safePropsTypeName = safeTypeName(propsTypeName, typeNameMapping);
  let output = `declare let properties: { children?: any } & ${argKeys.length === 0 ? safePropsTypeName : `__previewjs_Optionalize<${safePropsTypeName}, ${argKeys.map((prop) => `"${prop.replace(/"/g, '\\"')}"`).join(" | ")}>`};

type __previewjs_Optionalize<T, K extends keyof T> = {
  [P in Exclude<__previewjs_RequiredKeys<T>, K>]: T[P]
} & {
  [P in Exclude<__previewjs_OptionalKeys<T>, K>]?: T[P]
} & {
  [P in K]?: T[P]
}
type __previewjs_RequiredKeys<T> = { [K in keyof T]-?: {} extends Pick<T, K> ? never : K }[keyof T];
type __previewjs_OptionalKeys<T> = { [K in keyof T]-?: {} extends Pick<T, K> ? K : never }[keyof T];

declare function fn<T>(name: string, returnValue?: T): () => T;`;
  if (!hasProperType) {
    output += `

type ${safePropsTypeName} = ${generateTypeScriptType(type, collected, typeNameMapping, usedTypes, true)};`;
  }
  let generatedCount;
  do {
    generatedCount = 0;
    for (const typeName of [...usedTypes]) {
      if (declaredTypes.has(typeName)) {
        continue;
      }
      const type2 = collected[typeName];
      if (!type2) {
        continue;
      }
      let params = "";
      if (Object.keys(type2.parameters).length > 0) {
        params = "<";
        for (const [paramName, defaultType] of Object.entries(type2.parameters)) {
          params += paramName;
          if (defaultType) {
            params += ` = ${generateTypeScriptType(defaultType, collected, typeNameMapping, usedTypes)}`;
          }
          params += ",";
        }
        params += ">";
      }
      output += `

type ${safeTypeName(typeName, typeNameMapping)}${params} = ${generateTypeScriptType(type2.type, collected, typeNameMapping, usedTypes, typeName === propsTypeName)}`;
      declaredTypes.add(typeName);
      generatedCount += 1;
    }
  } while (generatedCount > 0);
  return output;
}
function generateTypeScriptType(type, collected, typeNameMapping, usedTypes, optionalFunctionProps = false) {
  switch (type.kind) {
    case "any":
      return "any";
    case "unknown":
      return "unknown";
    case "never":
      return "never";
    case "void":
      return "void";
    case "null":
      return "null";
    case "boolean":
      return "boolean";
    case "string":
      return "string";
    case "number":
      return "number";
    case "node":
      return "any";
    case "literal":
      if (typeof type.value === "number") {
        return type.value.toString(10);
      } else if (typeof type.value === "string") {
        return `"${type.value.replace(/"/g, '\\"')}"`;
      } else {
        return type.value ? "true" : "false";
      }
    case "enum":
      return Object.values(type.options).map((value) => typeof value === "number" ? value.toString(10) : `"${value.replace(/"/g, '\\"')}"`).join(" | ");
    case "array":
      return `Array<${generateTypeScriptType(type.items, collected, typeNameMapping, usedTypes)}>`;
    case "set":
      return `Set<${generateTypeScriptType(type.items, collected, typeNameMapping, usedTypes)}>`;
    case "object":
      return `{
              ${Object.entries(type.fields).filter(([propName]) => propName !== "children").map(([propName, propType]) => {
        const resolvedPropType = propType.kind === "name" ? collected[propType.name]?.type || propType : propType;
        const optional = optionalFunctionProps && resolvedPropType?.kind === "function" || resolvedPropType.kind === "optional" || resolvedPropType.kind === "any" || resolvedPropType.kind === "unknown";
        return `["${propName.replace(/"/g, '\\"')}"]${optional ? "?" : ""}: ${generateTypeScriptType(propType, collected, typeNameMapping, usedTypes)}`;
      }).join("\n")}
            }`;
    case "map":
      return `Map<${generateTypeScriptType(type.keys, collected, typeNameMapping, usedTypes)}, ${generateTypeScriptType(type.values, collected, typeNameMapping, usedTypes)}>`;
    case "record":
      return `Record<${generateTypeScriptType(type.keys, collected, typeNameMapping, usedTypes)}, ${generateTypeScriptType(type.values, collected, typeNameMapping, usedTypes)}>`;
    case "union":
      return type.types.map((subtype) => `(${generateTypeScriptType(subtype, collected, typeNameMapping, usedTypes)})`).join(" | ");
    case "intersection":
      return type.types.map((subtype) => `(${generateTypeScriptType(subtype, collected, typeNameMapping, usedTypes)})`).join(" & ");
    case "function":
      return `(...params: any[]) => (${generateTypeScriptType(type.returnType, collected, typeNameMapping, usedTypes)})`;
    case "optional":
      return `${generateTypeScriptType(type.type, collected, typeNameMapping, usedTypes)} | undefined`;
    case "promise":
      return `Promise<${generateTypeScriptType(type.type, collected, typeNameMapping, usedTypes)}>`;
    case "name":
      usedTypes.add(type.name);
      let name = safeTypeName(type.name, typeNameMapping);
      if (type.args.length > 0) {
        name += "<";
        for (let i = 0; i < type.args.length; i++) {
          if (i > 0) {
            name += ",";
          }
          const arg = type.args[i];
          name += generateTypeScriptType(arg, collected, typeNameMapping, usedTypes);
        }
        name += ">";
      }
      return name;
    default:
      throw assertNever$1(type);
  }
}
function safeTypeName(fullTypeName, typeNameMapping) {
  const existingName = typeNameMapping[fullTypeName];
  if (existingName) {
    return existingName;
  }
  const typeName = fullTypeName.split(":")[1];
  if (!typeName) {
    return fullTypeName;
  }
  let i = 1;
  const existingTypeNames = new Set(Object.values(typeNameMapping));
  while (existingTypeNames.has(numberedName(typeName, i))) {
    i++;
  }
  const result = numberedName(typeName, i);
  typeNameMapping[fullTypeName] = result;
  return result;
}
function numberedName(typeName, i) {
  return `${typeName}${i === 1 ? "" : `_${i}`}`;
}

class ComponentProps {
  constructor(name, types, argKeys, cachedInvocationSource) {
    this.name = name;
    this.types = types;
    this.argKeys = argKeys;
    this._invocationSource = cachedInvocationSource;
    makeAutoObservable(this);
  }
  setInvocationSource(source) {
    this._invocationSource = source;
  }
  get defaultPropsSource() {
    return generateDefaultProps(this.types.props, this.types.all);
  }
  get typeDeclarations() {
    return generateTypeDeclarations(this.name, this.types.props, this.argKeys, this.types.all);
  }
  get invocationSource() {
    const source = this._invocationSource;
    if (source === null) {
      return this.defaultInvocationSource;
    }
    return source;
  }
  get isDefaultInvocationSource() {
    return !this.invocationSource || this.invocationSource === this.defaultInvocationSource;
  }
  get defaultInvocationSource() {
    return generateInvocation(this.types.props, [
      ...extractFunctionKeys(this.types.props, this.types.all),
      ...this.argKeys
    ], this.types.all);
  }
}

const REFRESH_PERIOD_MILLIS = 5e3;
class PreviewState {
  constructor(localApi, webApi, persistedStateController, options = {}) {
    this.localApi = localApi;
    this.persistedStateController = persistedStateController;
    this.options = options;
    this.actionLogs = new ActionLogsState();
    this.consoleLogs = new ConsolePanelState();
    this.reachable = true;
    this.component = null;
    this.appInfo = null;
    this.iframeRef = {
      current: null
    };
    this.cachedInvocations = {};
    this.pingInterval = null;
    this.messageListener = (event) => {
      const data = event.data;
      if (data && data.kind === "navigate") {
        window.__previewjs_navigate(data.componentId);
      }
    };
    this.popStateListener = () => {
      this.onUrlChanged().catch(console.error);
    };
    this.keydownListener = (e) => {
      if (!navigator.userAgent.includes(" Code/")) {
        return;
      }
      const hasMeta = e.ctrlKey || e.metaKey;
      if (!hasMeta) {
        return;
      }
      switch (e.key.toLowerCase()) {
        case "c":
          document.execCommand("copy");
          e.preventDefault();
          break;
        case "x":
          document.execCommand("cut");
          e.preventDefault();
          break;
        case "v":
          document.execCommand("paste");
          e.preventDefault();
          break;
      }
    };
    this.updateBanner = new UpdateBannerState(webApi, this.persistedStateController);
    this.iframeController = createController({
      getIframe: () => this.iframeRef.current,
      listener: (event) => {
        runInAction(() => {
          if (!this.component?.details) {
            return;
          }
          switch (event.kind) {
            case "bootstrapped":
            case "before-render":
              this.consoleLogs.onClear();
              break;
            case "update":
              this.consoleLogs.onClear();
              if (event.rendering?.kind === "success") {
                this.component.variantKey = event.rendering.variantKey;
                this.component.details.variants = event.rendering.variants;
              }
              break;
            case "log-message":
              this.consoleLogs.onConsoleMessage(event);
              break;
            case "action":
              this.actionLogs.onAction(event);
              break;
            default:
              throw assertNever$1(event);
          }
        });
      }
    });
    makeAutoObservable(this, {
      iframeRef: observable.ref,
      pingInterval: observable.ref,
      appInfo: observable.ref
    });
  }
  async start() {
    window.__previewjs_navigate = (componentId, variantKey) => {
      history.pushState(null, "", `/?p=${componentId}${variantKey ? `&v=${variantKey}` : ""}`);
      this.onUrlChanged().catch(console.error);
    };
    window.addEventListener("message", this.messageListener);
    window.addEventListener("popstate", this.popStateListener);
    this.iframeController.start();
    await this.onUrlChanged();
    this.pingInterval = setInterval(() => {
      this.ping().catch(console.error);
    }, REFRESH_PERIOD_MILLIS);
    document.addEventListener("keydown", this.keydownListener);
    const { appInfo } = await this.localApi.request(localEndpoints.GetInfo);
    runInAction(() => {
      this.appInfo = appInfo;
    });
    await this.persistedStateController.start();
    await this.updateBanner.start(appInfo);
  }
  stop() {
    if (this.pingInterval) {
      clearInterval(this.pingInterval);
      this.pingInterval = null;
    }
    this.iframeController.stop();
    window.removeEventListener("message", this.messageListener);
    window.removeEventListener("popstate", this.popStateListener);
    document.removeEventListener("keydown", this.keydownListener);
  }
  setIframeRef(iframeRef) {
    if (iframeRef === this.iframeRef) {
      return;
    }
    this.iframeRef = iframeRef;
    this.renderComponent();
  }
  setComponent(componentId) {
    if (componentId === this.component?.componentId) {
      this.setVariant("custom");
    } else {
      window.__previewjs_navigate(componentId);
    }
  }
  setVariant(variantKey) {
    if (!this.component) {
      return;
    }
    window.__previewjs_navigate(this.component.componentId, variantKey);
  }
  updateProps(source) {
    if (!this.component?.details) {
      return;
    }
    this.component.details.props.setInvocationSource(source);
    this.cachedInvocations[this.component.componentId] = source;
    this.renderComponent();
  }
  resetProps() {
    if (!this.component?.details) {
      return;
    }
    this.component.details.props.setInvocationSource(null);
    delete this.cachedInvocations[this.component.componentId];
    this.renderComponent();
  }
  async onUrlChanged() {
    const urlParams = new URLSearchParams(document.location.search);
    const componentId = urlParams.get("p") || "";
    const variantKey = urlParams.get("v") || null;
    const decodedComponentId = decodeComponentId(componentId);
    if (this.options.onFileChanged) {
      await this.options.onFileChanged(decodedComponentId.currentFilePath);
    }
    if (!decodedComponentId.component) {
      this.component = null;
      return;
    }
    const name = decodedComponentId.component.name;
    if (this.component?.componentId === componentId) {
      runInAction(() => {
        if (this.component) {
          this.component.variantKey = variantKey;
        }
      });
    } else {
      this.iframeController.showLoading();
      runInAction(() => {
        this.component = {
          componentId,
          name,
          variantKey,
          details: null
        };
      });
      const response = await this.localApi.request(localEndpoints.ComputeProps, {
        filePath: decodedComponentId.component.filePath,
        componentName: name
      });
      const filePath = decodedComponentId.component.filePath;
      runInAction(() => {
        this.component = {
          componentId,
          name,
          variantKey,
          details: {
            filePath,
            variants: null,
            props: new ComponentProps(name, response.types, response.args, this.cachedInvocations[componentId] || null)
          }
        };
      });
    }
    this.renderComponent();
  }
  renderComponent() {
    if (!this.component?.details) {
      return;
    }
    this.consoleLogs.onClear();
    this.iframeController.loadComponent({
      componentName: this.component.name,
      filePath: this.component.details.filePath,
      variantKey: this.component.variantKey,
      customVariantPropsSource: this.component.details.props.invocationSource,
      defaultPropsSource: this.component.details.props.defaultPropsSource
    });
  }
  async ping() {
    let reachable;
    try {
      await fetch("/ping/");
      reachable = true;
    } catch (e) {
      console.warn(e);
      reachable = false;
    }
    runInAction(() => {
      this.reachable = reachable;
    });
  }
}

export { Link, PersistedStateController, Preview, PreviewState, Selection, createLocalApi, createWebApi };
